/*

   xlog - GTK+ logging program for amateur radio operators
   Copyright (C) 2001 - 2008 Joop Stakenborg <pg4i@amsat.org>

   This file is part of xlog.

   Xlog is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Xlog is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xlog.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <gtk/gtk.h>
#include <math.h>

#include "gui_gtkprintdialog.h"
#include "cfg.h"
#include "log.h"
#include "support.h"
#include "utils.h"

extern GtkWidget *mainwindow;
extern GtkWidget *mainnotebook;
extern GList *logwindowlist;
extern preferencestype preferences;

typedef struct 
{
	gdouble font_height;
	gdouble font_width;
	gdouble top_margin;
	gdouble left_margin;
	gdouble page_height;
	gdouble page_width;
	guint lines_per_page;
	guint num_pages;
} PrintData;
GtkPageSetup *print_page_setup;
GtkPrintSettings *print_settings;
static const char *printtype = "sans 12";

static void
end_print (GtkPrintOperation *operation, GtkPrintContext *context,
	gpointer user_data)
{
	PrintData *data = (PrintData *)user_data;

	g_free (data);
}

static void
begin_print (GtkPrintOperation *operation, GtkPrintContext *context,
	gpointer user_data)
{
	PrintData *data = (PrintData *)user_data;
	PangoFontDescription *desc;
	PangoFontMetrics *metrics;
	PangoLayout *layout;
	logtype *logwindow;

	if (!print_settings)
		print_settings = gtk_print_settings_new ();
	if (!print_page_setup)
		print_page_setup = gtk_page_setup_new ();
	gtk_print_operation_set_print_settings (operation, print_settings);
	gtk_print_operation_set_default_page_setup (operation, print_page_setup);

	logwindow = g_list_nth_data (logwindowlist,
		gtk_notebook_get_current_page (GTK_NOTEBOOK(mainnotebook)));

	GtkPageOrientation orientation =
		gtk_page_setup_get_orientation (print_page_setup);
	switch (orientation) {
		case GTK_PAGE_ORIENTATION_PORTRAIT:
		case GTK_PAGE_ORIENTATION_REVERSE_PORTRAIT:
		data->lines_per_page = 50;
		break;
		case GTK_PAGE_ORIENTATION_LANDSCAPE:
		case GTK_PAGE_ORIENTATION_REVERSE_LANDSCAPE:
		data->lines_per_page = 30;
		default:
		break;
	}

	data->num_pages = floor (1 + logwindow->qsos/data->lines_per_page);
	gtk_print_operation_set_n_pages (operation, data->num_pages);

	desc = pango_font_description_from_string (printtype);
	data->font_height =
		pango_units_to_double (pango_font_description_get_size (desc));

	layout = gtk_print_context_create_pango_layout (context);
	metrics = pango_context_get_metrics (pango_layout_get_context (layout),
		desc, pango_language_get_default ());
	data->font_width =
		pango_font_metrics_get_approximate_char_width (metrics)
		/ PANGO_SCALE;

	pango_font_description_free (desc);
	pango_font_metrics_unref (metrics);
	g_object_unref (layout);

	data->top_margin =
		gtk_page_setup_get_top_margin (print_page_setup, GTK_UNIT_POINTS);
	data->left_margin =
		gtk_page_setup_get_left_margin (print_page_setup, GTK_UNIT_POINTS);
	data->page_height =
		gtk_page_setup_get_page_height (print_page_setup, GTK_UNIT_POINTS);
	data->page_width =
		gtk_page_setup_get_page_width (print_page_setup, GTK_UNIT_POINTS);
}

gint
colwidth (gpointer arg, gint column, gint firstqso, gpointer user_data)
{
	PrintData *data = (PrintData *)user_data;
	gint i, lastqso, width = 0;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;
	gchar *pathstr, *it;

	lastqso = firstqso - data->lines_per_page;
	if (lastqso < 0) lastqso = 0;

	/* make sure the header fits */
	if (column == NR)
		width = g_utf8_strlen ("Nr", -1);
	else if (column == DATE)
		width = g_utf8_strlen ("Date", -1);
	else if (column == GMT)
		width = g_utf8_strlen ("UTC", -1);
	else if (column == GMTEND)
		width = g_utf8_strlen ("END", -1);
	else if (column == CALL)
		width = g_utf8_strlen ("Callsign", -1);
	else if (column == BAND)
		width = g_utf8_strlen ("MHz", -1);
	else if (column == MODE)
		width = g_utf8_strlen ("Mode", -1);
	else if (column == RST)
		width = g_utf8_strlen ("TX", -1);
	else if (column == MYRST)
		width = g_utf8_strlen ("RX", -1);
	else if (column == AWARDS)
		width = g_utf8_strlen ("Awards", -1);
	else if (column == QSLOUT)
		width = g_utf8_strlen ("Out", -1);
	else if (column == QSLIN)
		width = g_utf8_strlen ("In", -1);
	else if (column == POWER)
		width = g_utf8_strlen ("Pwr", -1);
	else if (column == NAME)
		width = g_utf8_strlen ("Name", -1);
	else if (column == QTH)
		width = g_utf8_strlen ("QTH", -1);
	else if (column == LOCATOR)
		width = g_utf8_strlen ("Locator", -1);
	else if (column == U1)
		width = g_utf8_strlen (preferences.freefield1, -1);
	else if (column == U2)
		width = g_utf8_strlen (preferences.freefield2, -1);
	else if (column == REMARKS)
		width = g_utf8_strlen ("Remarks", -1);
	else
		width = 4;

	logtype *logwindow = (logtype *) arg;
	model = gtk_tree_view_get_model (GTK_TREE_VIEW(logwindow->treeview));
	for (i = firstqso; i > lastqso; i--)
	{
		pathstr = g_strdup_printf("%d", i);
		path = gtk_tree_path_new_from_string (pathstr);
		if (path)
		{
			if (gtk_tree_model_get_iter (model, &iter, path))
			{
				gtk_tree_model_get (model, &iter, column, &it, -1);
				width = g_utf8_strlen (it, -1) > width
					? g_utf8_strlen (it, -1) : width;
			}
			else
				g_warning ("gui_gtkprintdialog->colwidth->GtkTreeIter");
		}
		else
			g_warning ("gui_gtkprintdialog->colwidth->GtkTreePath");
		g_free (pathstr);
	}
	if (column == CALL)
		return (width + 4) * data->font_width;
	else
		return (width + 3) * data->font_width;
}

gboolean
column_visible (gpointer arg, gint nr)
{
	logtype *logwindow = (logtype *) arg;
	GtkTreeViewColumn *column =
		gtk_tree_view_get_column (GTK_TREE_VIEW(logwindow->treeview), nr);
	return gtk_tree_view_column_get_visible (column);
}  


static void
draw_page (GtkPrintOperation *operation, GtkPrintContext *context,
	gint page_nr, gpointer user_data)
{
	logtype *logwindow;
	cairo_t *cr;
	PangoLayout *layout;
	PangoFontDescription *desc;
	gint i, j, left, linecount, first, last, top = 20;
	gint wnr = 0, wdate = 0, wgmt = 0, wgmtend = 0, wcall = 0,
		wband = 0, wmode = 0, wrst = 0, wmyrst = 0, wawards = 0,
		wqslout = 0, wqslin = 0, wpower = 0, wname = 0, wqth = 0,
		wlocator = 0, wu1 = 0, wu2 = 0, wremarks = 0;
	gboolean gmtendok = FALSE, awardsok = FALSE, qsloutok = FALSE,
		qslinok = FALSE, powerok = FALSE,nameok = FALSE, qthok = FALSE,
		locatorok = FALSE, u1ok = FALSE, u2ok = FALSE, remarksok = FALSE;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;
	gchar *pathstr, **item;


	PrintData *data = (PrintData *)user_data;
	cr = gtk_print_context_get_cairo_context (context);

	logwindow = g_list_nth_data (logwindowlist,
		gtk_notebook_get_current_page (GTK_NOTEBOOK(mainnotebook)));
	model = gtk_tree_view_get_model (GTK_TREE_VIEW (logwindow->treeview));
	if (gtk_tree_model_get_iter_first (model, &iter))
	{
		linecount = 0;
		left = data->left_margin;
		top = data->top_margin;

		layout = gtk_print_context_create_pango_layout (context);
		desc = pango_font_description_from_string (printtype);
		pango_font_description_set_weight (desc, PANGO_WEIGHT_BOLD);
		pango_layout_set_font_description (layout, desc);

		item = g_new0 (gchar *, QSO_FIELDS);

		/* calculate first and last QSO for this page,
		 * NB: page_nr starts at zero */
		first = logwindow->qsos - (page_nr * data->lines_per_page) - 1;
		last = first - data->lines_per_page;
		if (last < 0) last = 0;

		/* print the header and calculate column width for next 50 QSO's */
		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "Nr", -1);
		pango_cairo_show_layout (cr, layout);
		wnr = colwidth (logwindow, NR, first, data);
		left = left + wnr;

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "Date", -1);
		pango_cairo_show_layout (cr, layout);
		wdate = colwidth (logwindow, DATE, first, data);
		left = left + wdate;

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "UTC", -1);
		pango_cairo_show_layout (cr, layout);
		wgmt = colwidth (logwindow, GMT, first, data);
		left = left + wgmt;

		gmtendok = column_visible (logwindow, GMTEND);
		if (gmtendok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "END", -1);
			pango_cairo_show_layout (cr, layout);
			wgmtend = colwidth (logwindow, GMTEND, first, data);
			left = left + wgmtend;
		}

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "Callsign", -1);
		pango_cairo_show_layout (cr, layout);
		wcall = colwidth (logwindow, CALL, first, data);
		left = left + wcall;

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "MHz", -1);
		pango_cairo_show_layout (cr, layout);
		wband = colwidth (logwindow, BAND, first, data);
		left = left + wband;

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "Mode", -1);
		pango_cairo_show_layout (cr, layout);
		wmode = colwidth (logwindow, MODE, first, data);
		left = left + wmode;

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "TX", -1);
		pango_cairo_show_layout (cr, layout);
		wrst = colwidth (logwindow, RST, first, data);
		left = left + wrst;

		cairo_move_to (cr, left, top);
		pango_layout_set_text (layout, "RX", -1);
		pango_cairo_show_layout (cr, layout);
		wmyrst = colwidth (logwindow, MYRST, first, data);
		left = left + wmyrst;

		awardsok = column_visible (logwindow, AWARDS);
		if (awardsok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "Awards", -1);
			pango_cairo_show_layout (cr, layout);
			wawards = colwidth (logwindow, AWARDS, first, data);
			left = left + wawards;
		}

		qsloutok = column_visible (logwindow, QSLOUT);
		if (qsloutok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "Out", -1);
			pango_cairo_show_layout (cr, layout);
			wqslout = colwidth (logwindow, QSLOUT, first, data);
			left = left + wqslout;
		}

		qslinok = column_visible (logwindow, QSLIN);
		if (qslinok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "In", -1);
			pango_cairo_show_layout (cr, layout);
			wqslin = colwidth (logwindow, QSLIN, first, data);
			left = left + wqslin;
		}

		powerok = column_visible (logwindow, POWER);
		if (powerok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "Pwr", -1);
			pango_cairo_show_layout (cr, layout);
			wpower = colwidth (logwindow, POWER, first, data);
			left = left + wpower;
		}

		nameok = column_visible (logwindow, NAME);
		if (nameok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "Name", -1);
			pango_cairo_show_layout (cr, layout);
			wname = colwidth (logwindow, NAME, first, data);
			left = left + wname;
		}

		qthok = column_visible (logwindow, QTH);
		if (qthok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "QTH", -1);
			pango_cairo_show_layout (cr, layout);
			wqth = colwidth (logwindow, QTH, first, data);
			left = left + wqth;
		}

		locatorok = column_visible (logwindow, LOCATOR);
		if (locatorok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, "Locator", -1);
			pango_cairo_show_layout (cr, layout);
			wlocator = colwidth (logwindow, LOCATOR, first, data);
			left = left + wlocator;
		}

		u1ok = column_visible (logwindow, U1);
		if (u1ok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, preferences.freefield1, -1);
			pango_cairo_show_layout (cr, layout);
			wu1 = colwidth (logwindow, U1, first, data);
			left = left + wu1;
		}

		u2ok = column_visible (logwindow, U2);
		if (u2ok)
		{
			cairo_move_to (cr, left, top);
			pango_layout_set_text (layout, preferences.freefield2, -1);
			pango_cairo_show_layout (cr, layout);
			wu2 = colwidth (logwindow, U2, first, data);
			left = left + wu2;
		}

		remarksok = column_visible (logwindow, REMARKS);
		if (remarksok)
		{
			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, "Remarks", -1);
			pango_cairo_show_layout (cr, layout);
			wremarks = colwidth (logwindow, REMARKS, first, data);
			left = left + wremarks;
		}

		pango_font_description_set_weight (desc, PANGO_WEIGHT_NORMAL);
		pango_layout_set_font_description (layout, desc);

		linecount = pango_layout_get_line_count (layout);
		top = top + (0.5 + linecount) * data->font_height;
		left = data->left_margin;
		cairo_set_source_rgb (cr, 0, 0, 0);
		cairo_set_line_width (cr, 1);
		cairo_move_to (cr, left, top);
		cairo_line_to (cr, data->page_width, top);
		cairo_stroke (cr);
		linecount = pango_layout_get_line_count (layout);
		top = top + linecount * data->font_height;

		/* print QSO's */
		for (i = first; i > last; i--)
		{
			pathstr = g_strdup_printf("%d", i);
			path = gtk_tree_path_new_from_string (pathstr);
			if (path) {
			if (gtk_tree_model_get_iter (model, &iter, path)) {

			for (j = 0; j < QSO_FIELDS; j++)
				gtk_tree_model_get (model, &iter, j, &item[j], -1);

			left = data->left_margin;

			cairo_move_to (cr, left, top);
	  		pango_layout_set_text (layout, item[NR], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wnr;

			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, item[DATE], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wdate;

			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, item[GMT], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wgmt;

			if (gmtendok)
			{
				cairo_move_to (cr, left, top);
		  		pango_layout_set_text (layout, item[GMTEND], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wgmtend;
			}

			pango_font_description_set_weight (desc, PANGO_WEIGHT_BOLD);
			pango_layout_set_font_description (layout, desc);
			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, item[CALL], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wcall;
			pango_font_description_set_weight (desc, PANGO_WEIGHT_NORMAL);
			pango_layout_set_font_description (layout, desc);

			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, item[BAND], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wband;

			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, item[MODE], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wmode;

			cairo_move_to (cr, left, top);
		  	pango_layout_set_text (layout, item[RST], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wrst;

			cairo_move_to (cr, left, top);
	  		pango_layout_set_text (layout, item[MYRST], -1);
			pango_cairo_show_layout (cr, layout);
			left = left + wmyrst;

			if (awardsok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[AWARDS], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wawards;
			}

			if (qslinok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[QSLIN], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wqslin;
			}

			if (qsloutok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[QSLOUT], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wqslout;
			}

			if (powerok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[POWER], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wpower;
			}

			if (nameok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[NAME], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wname;
			}

			if (qthok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[QTH], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wqth;
			}

			if (locatorok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[LOCATOR], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wlocator;
			}

			if (u1ok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[U1], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wu1;
			}

			if (u2ok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[U2], -1);
				pango_cairo_show_layout (cr, layout);
				left = left + wu2;
			}

			if (remarksok)
			{
				cairo_move_to (cr, left, top);
			  	pango_layout_set_text (layout, item[REMARKS], -1);
				pango_cairo_show_layout (cr, layout);
			}

			linecount = pango_layout_get_line_count (layout);
			top = top + linecount * data->font_height;

			} else g_warning ("gui_gtkprintdialog->draw_page->GtkTreePath");
			} else g_warning ("gui_gtkprintdialog->draw_page->GtkTreeIter");

		}
		left = data->left_margin;
		top = data->page_height - 1.5 * data->font_height;
		cairo_set_source_rgb (cr, 0, 0, 0);
		cairo_set_line_width (cr, 1);
		cairo_move_to (cr, left, top);
		cairo_line_to (cr, data->page_width, top);
		cairo_stroke (cr);

		pango_font_description_set_style (desc, PANGO_STYLE_ITALIC);
		pango_layout_set_font_description (layout, desc);
		left = data->left_margin;
		top = data->page_height - 1 * data->font_height;
		cairo_move_to (cr, left, top);
		gchar *temp = g_strdup_printf
			(_("Log %s printed on %s by %s version %s"),
			logwindow->logname, xloggetdate(), PACKAGE, VERSION);
	  	pango_layout_set_text (layout, temp, -1);
	  	pango_layout_set_alignment (layout, PANGO_ALIGN_CENTER);
	  	pango_layout_set_width
	  		(layout, pango_units_from_double (data->page_width));
		pango_cairo_show_layout (cr, layout);

		g_free (temp);
		pango_font_description_free (desc);
		g_object_unref (layout);
	}
}

void
on_menu_print_activate (GtkMenuItem * menuitem, gpointer user_data)
{
	GtkPrintOperation *operation;
	GError *error = NULL;
	PrintData *data;

	operation = gtk_print_operation_new ();
	data = g_new0 (PrintData, 1);

	g_signal_connect (G_OBJECT (operation), "begin-print", 
				G_CALLBACK (begin_print), data);
	g_signal_connect (G_OBJECT (operation), "draw-page", 
				G_CALLBACK (draw_page), data);
	g_signal_connect (G_OBJECT (operation), "end-print", 
				G_CALLBACK (end_print), data);

	gtk_print_operation_run (operation, GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG,
		GTK_WINDOW (mainwindow), &error);

	g_object_unref (operation);
	if (error)
	{
		GtkWidget *dialog;

		dialog = gtk_message_dialog_new (GTK_WINDOW (mainwindow),
			GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
			"%s", error->message);
		g_error_free (error);
		g_signal_connect (dialog, "response", G_CALLBACK (gtk_widget_destroy), NULL);
		gtk_widget_show (dialog); 
	}

}
